#! /usr/bin/env python3

"""PDF Quality Assessment for Digitisation batches

Johan van der Knijff

Copyright 2024, KB/National Library of the Netherlands

Module with shared functions

"""

import sys
import os

def errorExit(msg):
    """Write error to stderr and exit"""
    msgString = "ERROR: {}\n".format(msg)
    sys.stderr.write(msgString)
    sys.exit()


def checkFileExists(fileIn):
    """Check if file exists and exit if not"""
    if not os.path.isfile(fileIn):
        msg = "file {} does not exist".format(fileIn)
        errorExit(msg)


def checkDirExists(pathIn):
    """Check if directory exists and exit if not"""
    if not os.path.isdir(pathIn):
        msg = "directory {} does not exist".format(pathIn)
        errorExit(msg)



#! /usr/bin/env python3

"""Image Quality Assessment for Digitisation batches

Johan van der Knijff

Copyright 2025, KB/National Library of the Netherlands

"""

import sys
import os
import shutil
import time
import argparse
import csv
import logging
from lxml import etree
from . import properties
from . import schematron
from . import shared

__version__ = "0.1.6"

# Create parser
parser = argparse.ArgumentParser(description="IMaGe QUality Assessment for Digitisation batches")


def parseCommandLine():
    """Parse command line"""

    # Sub-parsers for process and list commands

    subparsers = parser.add_subparsers(help='sub-command help',
                                       dest='subcommand')
    parser_process = subparsers.add_parser('process',
                                          help='process a batch')
    parser_process.add_argument('profile',
                                action="store",
                                help='validation profile name (use "imgquad list" to list available profiles)')
    parser_process.add_argument('batchDir',
                                action="store",
                                help="batch directory")
    parser_process.add_argument('--prefixout', '-p',
                                action="store",
                                default='iq',
                                help="prefix of output files")
    parser_process.add_argument('--outdir', '-o',
                                action="store",
                                default=os.getcwd(),
                                help="output directory")
    parser_process.add_argument('--delimiter', '-d',
                            action="store",
                            default=';',
                            help="output delimiter")
    parser_process.add_argument('--verbose', '-b',
                                action="store_true",
                                default=False,
                                help="report Schematron report in verbose format")
    parser_list = subparsers.add_parser('list',
                                        help='list available profiles and schemas')
    parser_copyps = subparsers.add_parser('copyps',
                                        help='copy default profiles and schemas to \
                                            user directory, note that this will overwrite \
                                            any user-modified versions of these files!')
    parser.add_argument('--version', '-v',
                        action="version",
                        version=__version__)

    # Parse arguments
    args = parser.parse_args()

    return args


def getFilesFromTree(rootDir, extensions):
    """Walk down whole directory tree (including all subdirectories) and
    return list of those files whose extensions match extensions list
    NOTE: directory names are disabled here!!
    implementation is case insensitive (all search items converted to
    upper case internally!
    """

    # Convert extensions to uppercase
    extensions = [extension.upper() for extension in extensions]
    filesList = []

    for dirname, dirnames, filenames in os.walk(rootDir):
        # Suppress directory names
        for subdirname in dirnames:
            thisDirectory = os.path.join(dirname, subdirname)

        for filename in filenames:
            thisFile = os.path.join(dirname, filename)
            thisExtension = os.path.splitext(thisFile)[1]
            thisExtension = thisExtension.upper().strip('.')
            if extensions[0].strip() == '*' or thisExtension in extensions:
                filesList.append(thisFile)
    return filesList


def writeXMLHeader(fileOut):
    """Write XML header"""
    xmlHead = "<?xml version='1.0' encoding='UTF-8'?>\n"
    xmlHead += "<imgquad>\n"
    with open(fileOut,"wb") as f:
        f.write(xmlHead.encode('utf-8'))


def writeXMLFooter(fileOut):
    """Write XML footer"""
    xmlFoot = "</imgquad>\n"
    with open(fileOut,"ab") as f:
        f.write(xmlFoot.encode('utf-8'))


def processFile(file, verboseFlag, schemas):
    """Process one file"""

    # Create output element for this file
    fileElt = etree.Element("file")

    # Initial value of flag that indicates whether image passes or fails quality checks
    validationOutcome = "Pass"
    # Initial value of flag that indicates whether validation was successful
    validationSuccess = False

    # Select schema based on directory or file name pattern defined in profile
    schemaMatchFlag, mySchema = schematron.findSchema(file, schemas)
    
    # Extract properties
    propertiesElt = properties.getProperties(file)

    # Validate extracted properties against schema
    if schemaMatchFlag:
        validationSuccess, validationOutcome, reportElt = schematron.validate(mySchema,
                                                                              propertiesElt,
                                                                              verboseFlag)
    else:
        # No schema match
        validationOutcome = "Fail"
        logging.warning("no schema match")

    if not validationSuccess:
        logging.warning("Schematron validation was not successful")

    # Create schema and status elements
    schemaElt = etree.Element("schema")
    schemaElt.text = mySchema
    validationSuccessElt = etree.Element("validationSuccess")
    validationSuccessElt.text = str(validationSuccess)
    validationOutcomeElt = etree.Element("validationOutcome")
    validationOutcomeElt.text = validationOutcome
    # Add all child elements to file element
    fileElt.append(propertiesElt)
    fileElt.append(schemaElt)
    fileElt.append(validationSuccessElt)
    fileElt.append(validationOutcomeElt)
    if schemaMatchFlag:
        fileElt.append(reportElt)

    return fileElt


def findEltValue(element, path, ns):
    """ Return text of path in element, or "n/a" if it doesn't exist """
    try:
        elOut = element.xpath(path, namespaces=ns)
        
        if len(elOut) > 0:
            if type(elOut[0]) == etree._Element:
                result = elOut[0].text
            elif type(elOut[0]) == etree._ElementUnicodeResult:
                result = elOut[0]
        else:
            result = "n/a"

    except Exception:
        raise
        result = "n/a"
    
    return result


def main():
    """Main function"""

    # Path to configuration dir (from https://stackoverflow.com/a/53222876/1209004
    # and https://stackoverflow.com/a/13184486/1209004).
    # TODO on Windows this should return the AppData/Local folder, does this work??
    configpath = os.path.join(
    os.environ.get('LOCALAPPDATA') or
    os.environ.get('XDG_CONFIG_HOME') or
    os.path.join(os.environ['HOME'], '.config'),
    "imgquad")

     # Create config directory if it doesn't exist already
    if not os.path.isdir(configpath):
        os.mkdir(configpath)
   
    # Locate package directory
    packageDir = os.path.dirname(os.path.abspath(__file__))

    # Profile and schema locations in installed package and config folder
    profilesDirPackage = os.path.join(packageDir, "profiles")
    schemasDirPackage = os.path.join(packageDir, "schemas")
    profilesDir = os.path.join(configpath, "profiles")
    schemasDir = os.path.join(configpath, "schemas")

    # Check if package profiles and schemas dirs exist
    shared.checkDirExists(profilesDirPackage)
    shared.checkDirExists(schemasDirPackage)

    # Copy profiles and schemas to respective dirs in config dir
    if not os.path.isdir(profilesDir):
        shutil.copytree(profilesDirPackage, profilesDir)
    if not os.path.isdir(schemasDir):
        shutil.copytree(schemasDirPackage, schemasDir)

    # Get input from command line
    args = parseCommandLine()
    action = args.subcommand

    if action == "process":
        # Check if all profiles and schemas can be parsed
        schematron.checkProfilesSchemas(profilesDir, schemasDir)
        profile = os.path.basename(args.profile)
        batchDir = os.path.normpath(args.batchDir)
        prefixOut = args.prefixout
        outDir = os.path.normpath(args.outdir)
        delimiter = args.delimiter
        verboseFlag = args.verbose
    elif action == "list":
        schematron.listProfilesSchemas(profilesDir, schemasDir)
    elif action == "copyps":
        shutil.copytree(profilesDirPackage, profilesDir, dirs_exist_ok=True)
        msg = ("copied profiles from {} to {}").format(profilesDirPackage, profilesDir)
        print(msg)
        shutil.copytree(schemasDirPackage, schemasDir, dirs_exist_ok=True)
        msg = ("copied schemas from {} to {}").format(schemasDirPackage, schemasDir)
        print(msg)
        sys.exit()
    elif action is None:
        print('')
        parser.print_help()
        sys.exit()
    
    # Add profilesDir to profile definition
    profile = os.path.join(profilesDir, profile)

    # Check if files / directories exist
    shared.checkFileExists(profile)
    shared.checkDirExists(batchDir)
    shared.checkDirExists(outDir)

    # Check if outDir is writable
    if not os.access(outDir, os.W_OK):
        msg = ("directory {} is not writable".format(outDir))
        shared.errorExit(msg)

    # Batch dir name
    batchDirName = os.path.basename(batchDir)
    # Construct output prefix for this batch
    prefixBatch = ("{}_{}").format(prefixOut, batchDirName)
    
    # Set up logging
    logging.basicConfig(handlers=[logging.StreamHandler(sys.stdout)],
                        level=logging.INFO,
                        format='%(asctime)s - %(levelname)s - %(message)s')

    # Get file extensions, summary properties schema patterns and locations from profile
    extensions, namespaces, summaryProperties, schemas = schematron.readProfile(profile, schemasDir)

    # Add Schematron namespace definition
    namespaces["svrl"] = "http://purl.oclc.org/dsdl/svrl"

    if len(extensions) == 0:
        msg = ("no file extensions defined in profile")
        shared.errorExit(msg)

    # Summary file with quality check status (pass/fail) and properties that are selected in profile
    summaryFile = os.path.normpath(("{}_summary.csv").format(prefixBatch))
    summaryFile = os.path.join(outDir, summaryFile)

    # List with names of output properties
    propertyNames = []
    for property in summaryProperties:
        propertyName = property.split('/')[-1]
        propertyNames.append(propertyName)

    summaryHeadings = ["file", "validationSuccess", "validationOutcome", "validationErrors"] + propertyNames

    with open(summaryFile, 'w', newline='', encoding='utf-8') as fSum:
        writer = csv.writer(fSum, delimiter=delimiter)
        writer.writerow(summaryHeadings)

    listFiles = getFilesFromTree(batchDir, extensions)
    # TODO: perhaps define extensions in profile?

    # start clock for statistics
    start = time.time()
    print("imgquad started: " + time.asctime())

    # Iterate over all files
    fileOut = ("{}.xml").format(prefixBatch)
    fileOut = os.path.join(outDir, fileOut)
    writeXMLHeader(fileOut)

    for myFile in listFiles:
        logging.info(("file: {}").format(myFile))
        myFile = os.path.abspath(myFile)
        fileResult = processFile(myFile, verboseFlag, schemas)
        if len(fileResult) != 0:
            validationSuccess = findEltValue(fileResult, 'validationSuccess', namespaces)
            validationOutcome = findEltValue(fileResult, 'validationOutcome', namespaces)
            with open(summaryFile, 'a', newline='', encoding='utf-8') as fSum:
                propValues = []
                for property in summaryProperties:
                    propertyValue = findEltValue(fileResult, property, namespaces)
                    propValues.append(propertyValue)

                validationErrors = []
                
                failedAsserts = fileResult.xpath("schematronReport/svrl:schematron-output/svrl:failed-assert/svrl:text", namespaces=namespaces)
                for failedAssert in failedAsserts:
                    validationErrors.append(failedAssert.text)
                validationErrorsString = '|'.join(validationErrors)
                
                writer = csv.writer(fSum, delimiter=delimiter)
                myRow = [myFile, validationSuccess, validationOutcome, validationErrorsString] + propValues
                writer.writerow(myRow)
            # Convert output to XML and add to output file
            outXML = etree.tostring(fileResult,
                                    method='xml',
                                    encoding='utf-8',
                                    xml_declaration=False,
                                    pretty_print=True)

            with open(fileOut,"ab") as f:
                f.write(outXML)

    writeXMLFooter(fileOut)

    # Timing output
    end = time.time()

    print("imgquad ended: " + time.asctime())

    # Elapsed time (seconds)
    timeElapsed = end - start
    timeInMinutes = round((timeElapsed / 60), 2)

    print("Elapsed time: {} minutes".format(timeInMinutes))


if __name__ == "__main__":
    main()



#! /usr/bin/env python3

"""ImagGe Quality Assessment for Digitisation batches

Johan van der Knijff

Copyright 2025, KB/National Library of the Netherlands

Image properties extraction module

"""
import os
import sys #remove, test only
import io
import logging
import base64
from lxml import etree
import PIL
from PIL import ImageCms
from PIL.TiffTags import TAGS as TAGS_TIFF
from PIL.ExifTags import TAGS as TAGS_EXIF, GPSTAGS, IFD
from . import jpegquality

def dictionaryToElt(name, dictionary):
    """Create Element object from dictionary, with recursion"""
    elt = etree.Element(name)

    for k, v in dictionary.items():
        if isinstance(v, dict):
            child = dictionaryToElt(str(k),v)
            elt.append(child)
        else:
            child = etree.Element(k)
            child.text = str(v)
        elt.append(child)

    return elt


def getBPC(image):
    """Return Bits per Component as a function of mode and components values"""
    mode_to_bpp = {"1": 1,
                   "L": 8,
                   "P": 8,
                   "RGB": 24,
                   "RGBA": 32,
                   "CMYK": 32,
                   "YCbCr": 24,
                   "LAB": 24,
                   "HSV": 24,
                   "I": 32,
                   "F": 32}

    bitsPerPixel = mode_to_bpp[image.mode]
    noComponents = len(image.getbands())

    if noComponents != 0  and isinstance(bitsPerPixel, int):
        bpc = int(bitsPerPixel/noComponents)
    else:
        bpc = -9999

    return bpc


def getProperties(file):
    """Extract properties and return result as Element object"""

    # Create element object to store all properties
    propertiesElt = etree.Element("properties")

    # Element to store exceptions at file level
    exceptionsFileElt = etree.Element("exceptions")

    # Create and fill descriptive elements
    fPathElt = etree.Element("filePath")
    fPathElt.text = file
    fNameElt = etree.Element("fileName")
    fNameElt.text = os.path.basename(file)
    fSizeElt = etree.Element("fileSize")
    fSizeElt.text = str(os.path.getsize(file))

    # Add to properies element
    propertiesElt.append(fPathElt)
    propertiesElt.append(fNameElt)
    propertiesElt.append(fSizeElt)

    # Read image
    try:
        im = PIL.Image.open(file)
        im.load()
        propsImageElt = getImageProperties(im)
        propertiesElt.append(propsImageElt)

    except Exception  as e:
        ex = etree.SubElement(exceptionsFileElt,'exception')
        ex.text = str(e)
        propertiesElt.append(exceptionsFileElt)
        logging.warning(("while opening image: {}").format(str(e)))
        raise
        return propertiesElt

    return propertiesElt


def getImageProperties(image):
    """Extract image properties and return result as Element object"""

    # Dictionary for storing image properties
    propsImage = {}
    # Element for storing image-level exceptions
    exceptionsImageElt = etree.Element("exceptions")

    propsImage['format'] = image.format
    width = image.size[0]
    height = image.size[1]
    propsImage['width'] = width
    propsImage['height'] = height
    propsImage['mode'] = image.mode
    noComponents = len(image.getbands())
    propsImage['components']= noComponents
    bitsPerComponent = getBPC(image)
    propsImage['bpc'] = bitsPerComponent

    if image.format == "JPEG":
        try:
            # Estimate JPEG quality using least squares matching
            # against standard quantization tables
            quality, rmsError, nse = jpegquality.computeJPEGQuality(image)
            propsImage['JPEGQuality'] = quality
            propsImage['NSE_JPEGQuality'] = nse
        except Exception as e:
            ex = etree.SubElement(exceptionsImageElt,'exception')
            ex.text = str(e)
            logging.warning(("while estimating JPEG quality from image: {}").format(str(e)))


    for key, value in image.info.items():

        if key == 'exif':
            # Skip any exif elements as Exif tags are added later
            pass
        elif key == 'photoshop':
            # Skip photoshop elements, because they tend to be large and I don't know how to
            # properly decode them
            pass
        elif isinstance(value, bytes):
            propsImage[key] = 'bytestream'
        elif key == 'dpi' and isinstance(value, tuple):
            propsImage['ppi_x'] = value[0]
            propsImage['ppi_y'] = value[1]
        elif key == 'jfif_density' and isinstance(value, tuple):
            propsImage['jfif_density_x'] = value[0]
            propsImage['jfif_density_y'] = value[1]
        elif isinstance(value, tuple):
            # Skip any other properties that return tuples
            pass
        else:
            propsImage[key] = value

    # ICC profile name and description
    iccFlag = False
    try:
        icc = image.info['icc_profile']
        iccFlag = True
    except KeyError:
        pass

    if iccFlag:
        try:
            iccProfile = ImageCms.ImageCmsProfile(io.BytesIO(icc))
            propsImage['icc_profile_name'] = ImageCms.getProfileName(iccProfile).strip()
            propsImage['icc_profile_description'] = ImageCms.getProfileDescription(iccProfile).strip()
        except Exception as e:
            ex = etree.SubElement(exceptionsImageElt,'exception')
            ex.text = str(e)
            logging.warning(("while extracting ICC profile properties from image: {}").format(str(e)))


    if image.format == "TIFF":
        # Create element object to store TIFF tags
        propsTIFFElt = etree.Element("tiff")

        # Iterate over TIFF tags, code adapted from:
        # https://stackoverflow.com/a/75357594/1209004 and
        # https://stackoverflow.com/a/46910779

        propsTIFF = {TAGS_TIFF[key] : image.tag[key] for key in image.tag.keys()}

        for k, d in propsTIFF.items():
            tag = k
            tiffElt = etree.Element(str(tag))

            # Don't include values of below tags
            if tag not in ['PhotoshopInfo', 'ICCProfile', 'IptcNaaInfo', 'XMP', 'ImageSourceData'] and isinstance(d, tuple):
                # extracted value is tuple, so reformat as spece-delimited string
                v = ''
                if tag not in ['XResolution', 'YResolution']:
                    for x in d:
                        v = v + ' ' + str(x)
                else:
                    try:
                        # In case of XResolution / YResolution tag, parse numerator and denominator
                        # values, and convert to resolution value
                        num = d[0][0]
                        den = d[0][1]
                        v = str(num/den)
                    except exception:
                        raise
                        pass

                tiffElt.text = v.strip()
            propsTIFFElt.append(tiffElt)

    # Exif tags
    propsExif = image.getexif()
    propsExifElt = etree.Element("exif")

    # Iterate over various Exif tags, code adapted from:
    # https://stackoverflow.com/a/75357594/1209004

    for k, v in propsExif.items():
        try:
            # This exception handler deals with any tags that Pillow doesn't recognize
            tag = TAGS_EXIF.get(k, k)
            exifElt = etree.Element(str(tag))
            if tag not in ['XMLPacket', 'InterColorProfile', 'IPTCNAA', 'ImageResources']:
                # Don't include content of these tags as text
                exifElt.text = str(v)

            propsExifElt.append(exifElt)
        except ValueError:
            pass

    for ifd_id in IFD:
        # Iterate over image file directories
        # NOTE: this can result in duplicate Exif Tags. Example: Thumbnail image is implemented as 
        # separate IFD, with XResolution / YResolution tags whose values are different from
        # main resolution tags. Currently these are all lumped together in the output.
        try:
            ifd = propsExif.get_ifd(ifd_id)

            if ifd_id == IFD.GPSInfo:
                resolve = GPSTAGS
            else:
                resolve = TAGS_EXIF

            for k, v in ifd.items():
                tag = resolve.get(k, k)
                exifElt = etree.Element(str(tag))
                exifElt.text = str(v)
                propsExifElt.append(exifElt)
        except KeyError:
            pass
    
    # Read XMP metadata as string since dedicated getxmp function returns dictionary
    # that is difficult to work with for our purposes 
    # See: https://github.com/python-pillow/Pillow/issues/5076#issuecomment-2119966091
    # this only works for TIFF!
    containsXMP = False
    if image.format == "TIFF":
        try:
            xmp = image.tag_v2[700].decode('utf-8')
            # Convert to Element object
            propsXMPElt = etree.fromstring(xmp)
            containsXMP = True
        except KeyError:
            pass

    propsImageElt = dictionaryToElt('image', propsImage)
    if image.format == "TIFF":
        propsImageElt.append(propsTIFFElt)
    propsImageElt.append(propsExifElt)
    if containsXMP:
        propsImageElt.append(propsXMPElt)
    propsImageElt.append(exceptionsImageElt)

    return propsImageElt



#! /usr/bin/env python3

"""PDF Quality Assessment for Digitisation batches

Johan van der Knijff

Copyright 2024, KB/National Library of the Netherlands

Module with code related to schematron, schemas and profiles

"""

import sys
import os
import logging
from lxml import isoschematron
from lxml import etree
from . import shared


def listProfilesSchemas(profilesDir, schemasDir):
    """List all available profiles and schemas"""
    profiles = os.listdir(profilesDir)
    print("Available profiles (directory {}):".format(profilesDir))
    for profile in profiles:
        print("  - {}".format(profile))
    schemas = os.listdir(schemasDir)
    print("Available schemas (directory {}):".format(schemasDir))
    for schema in schemas:
        print("  - {}".format(schema))
    sys.exit()


def checkProfilesSchemas(profilesDir, schemasDir):
    """Check if all profiles and schemas can be read without
    throwing parse errors"""
    profiles = os.listdir(profilesDir)
    for profile in profiles:
        try:
            readAsLXMLElt(os.path.join(profilesDir, profile))
        except Exception:
            msg = ("error parsing profile {}").format(profile)
            shared.errorExit(msg)
    schemas = os.listdir(schemasDir)
    for schema in schemas:
        try:
            schemaElt = readAsLXMLElt(os.path.join(schemasDir, schema))
        except Exception:
            msg = ("error parsing schema {}").format(schema)
            raise
            shared.errorExit(msg)
        try:
            isoschematron.Schematron(schemaElt)
        except etree.XSLTParseError:
            msg = ("XSLT parse error for schema {}").format(schema)
            raise
            shared.errorExit(msg)       


def readProfile(profile, schemasDir):
    """Read a profile and returns list with for each schema
    element the corresponding type, matching method, matching
    pattern and schematronj file"""

    # Parse XML tree
    try:
        tree = etree.parse(profile)
        prof = tree.getroot()
    except Exception:
        msg = "error parsing {}".format(profile)
        shared.errorExit(msg)

    # Output extensions list
    listExtensions = []

    # Output namespaces dictionary
    dictNamespaces = {}

    # Output properties list
    listProperties = []

    # Output schemas list
    listSchemas = []

    # Locate extension elements
    extensions = prof.findall("extension")

    # Add extensions to output list
    for extension in extensions:
        listExtensions.append(extension.text)

    # Locate namespace elements
    namespaces = prof.findall("ns")

    # Add namespace prefixes and uris to dictionary
    for namespace in namespaces:
        uri = namespace.attrib['uri']
        prefix = namespace.attrib['prefix']
        dictNamespaces[prefix] = uri

    # Locate summary properties elements and add them to list
    sProperties = prof.findall("summaryProperty")

    for property in sProperties:
        listProperties.append(property.text)

    # Flag that indicates use of "type" attribute
    hasType = True

    # Locate schema elements
    schemas = prof.findall("schema")

    # Add schemas to output list
    for schema in schemas:
        try:
            mType = schema.attrib["type"]
            if mType not in ["fileName", "parentDirName"]:
                msg = "'{}' is not a valid 'type' value".format(mType)
                shared.errorExit(msg)
        except KeyError:
            hasType = False

        if hasType:
            try:
                mMatch = schema.attrib["match"]
                if mMatch not in ["is", "startswith", "endswith", "contains"]:
                    msg = "'{}' is not a valid 'match' value".format(mMatch)
                    shared.errorExit(msg)
            except KeyError:
                msg = "missing 'match' attribute in profile {}".format(profile)
                shared.errorExit(msg)
            try:
                mPattern = schema.attrib["pattern"]
            except KeyError:
                msg = "missing 'pattern' attribute in profile {}".format(profile)
                shared.errorExit(msg)
        else:
            mType = None
            mMatch = None
            mPattern = None

        schematronFile = os.path.join(schemasDir, schema.text)
        shared.checkFileExists(schematronFile)

        listSchemas.append([mType, mMatch, mPattern, schematronFile])

    return listExtensions, dictNamespaces, listProperties, listSchemas


def readAsLXMLElt(xmlFile):
    """Parse XML file with lxml and return result as element object
    (not the same as Elementtree object!)
    """

    f = open(xmlFile, 'r', encoding="utf-8")
    # Note we're using lxml.etree here rather than elementtree
    resultAsLXMLElt = etree.parse(f)
    f.close()

    return resultAsLXMLElt


def summariseSchematron(report):
    """Return summarized version of Schematron report with only output of
    failed tests"""

    for elem in report.iter():
        if elem.tag == "{http://purl.oclc.org/dsdl/svrl}fired-rule":
            elem.getparent().remove(elem)

    return report


def findSchema(PDF, schemas):
    """Find schema based on match with name or parent directory"""

    # Initial value of flag that indicates schema match
    schemaMatchFlag = False
    # Initial value of schema reference
    schemaMatch = "undefined"

    fPath, fName = os.path.split(PDF)
    parentDir = os.path.basename(fPath)

    for schema in schemas:
        mType = schema[0]
        mMatch = schema[1]
        mPattern = schema[2]
        mSchema = schema[3]
        if mType == None:
            schemaMatch = mSchema
            schemaMatchFlag = True
        if mType == "parentDirName" and mMatch == "is":
            if parentDir == mPattern:
                schemaMatch = mSchema
                schemaMatchFlag = True
        elif mType == "parentDirName" and mMatch == "startswith":
            if parentDir.startswith(mPattern):
                schemaMatch = mSchema
                schemaMatchFlag = True
        elif mType == "parentDirName" and mMatch == "endswith":
            if parentDir.endswith(mPattern):
                schemaMatch = mSchema
                schemaMatchFlag = True
        elif mType == "parentDirName" and mMatch == "contains":
            if mPattern in parentDir:
                schemaMatch = mSchema
                schemaMatchFlag = True
        if mType == "fileName" and mMatch == "is":
            if fName == mPattern:
                schemaMatch = mSchema
                schemaMatchFlag = True
        elif mType == "fileName" and mMatch == "startswith":
            if fName.startswith(mPattern):
                schemaMatch = mSchema
                schemaMatchFlag = True
        elif mType == "fileName" and mMatch == "endswith":
            if fName.endswith(mPattern):
                schemaMatch = mSchema
                schemaMatchFlag = True
        elif mType == "fileName" and mMatch == "contains":
            if mPattern in fName:
                schemaMatch = mSchema
                schemaMatchFlag = True

    return schemaMatchFlag, schemaMatch


def validate(schema, propertiesElt, verboseFlag):
    """Validate extracted properties against schema"""

    # Initial value of validation outcome
    validationOutcome = "Pass"

    # Initial value of flag that indicates whether validation ran
    validationSuccess = False

    # Element used to store validation report
    reportElt = etree.Element("schematronReport")
    # Get schema as lxml.etree element
    mySchemaElt = readAsLXMLElt(schema)
    # Start Schematron magic ...
    schematron = isoschematron.Schematron(mySchemaElt,
                                          store_report=True)

    try:
        # Validate properties element against schema
        validationResult = schematron.validate(propertiesElt)
        # Set status to "Fail" if properties didn't pass validation
        if not validationResult:
            validationOutcome = "Fail"
        report = schematron.validation_report
        validationSuccess = True

    except Exception:
        validationOutcome = "Fail"
        logging.error(("Schematron validation failed for {}").format(schema))

    try:
        # Re-parse Schematron report
        report = etree.fromstring(str(report))
        # Make report less verbose
        if not verboseFlag:
            report = summariseSchematron(report)
        # Add to report element
        reportElt.append(report)
    except Exception:
        # No report available because Schematron validation failed
        pass

    return validationSuccess, validationOutcome, reportElt
